/* Status Bar at bottom */

#status-box {
	padding-top: 4px;
	padding-bottom: 3px;
}

/* Find, replace and jump dialogs */

#findreplace-buttonbox {
	padding-top: 6px;
}

#findreplace-box {
	background: @popover_bg_color;
	color: @popover_fg_color;
	padding: 6px;
	border-bottom: @borders solid 1px;
}

.findreplace-label {
    padding-left: 12px;
    padding-bottom: 2px;
    font-size: small;
}

.findreplace-frame hexwidget {
  border-top: shade(@theme_bg_color, 0.9) 1px dashed;
  border-bottom: shade(@theme_bg_color, 0.9) 1px dashed;
}

jumpdialog {
	padding-left: 12px;
}

jumpdialog entry {
	font-family: monospace;
}

markdialog {
	padding-left: 12px;
}

#mark-description-label {
	font-weight: lighter;
	font-family: monospace;
}

/* Conversions pane/box */

#conversions-box {
	padding: 12px;
}

/* Character table */

#chartable-action-area {
	padding: 6px;
}

/* Base converter */

#converter-grid {
	padding: 6px;
}

/* Info bar */

#info-bar {
	background-color: var(--shade-color);
	padding-left: 6px;
	padding-top: 6px;
	padding-bottom: 6px;
}
